{ A+,B-,D-,L-,M-,O+,Q-,R-,S-,X+,Y-}
(*
  SAR.PAS 1.1
  --------  --------  --------  --------  --------  --------  --------  --------
  Contents:
     TSearchAndReplace - Simple-To-Use Text Search & Replace component which
                         wraps the functionallity of TFindDialog and
                         TReplaceDialog.

  --------  --------  --------  --------  --------  --------  --------  --------
     Copyrights (C) 1997 ved T Skovmand Eriksen
     email: TSEriksen@cyberdude.com
     ALL RIGHTS RESERVED

     PARTS MAY BE COPYRIGHTED BY BORLAND INTL.
  --------  --------  --------  --------  --------  --------  --------  --------

  This source is freeware and may be used as follows:

    * In development of in-house applications which will not be published.

    * In development of freeware applications PROVIDED that credit is given
      to me in the applications aboutbox or helpfile.

    * In development of shareware and commercial applications PROVIDED that
      I, Troels S Eriksen, recieve a free unlimited copy of all versions of
      said application AND that credits are given to me in the applications
      aboutbox or helpfile.

    * In printed or electronic form (fx. in a book) or in a source-library
      (fx. on CD) PROVIDED that I, Troels S Eriksen, recieve a royality-free
      unlimited copy of said book or library.

    * In any other way PROVIDED that I, Troels S Eriksen, give my explicit
      permission to the suggested use.

  By using this source you do acknowledge that the above restrictions apply to
  you and your use of this component and that you will confirm by them.

  --------  --------  --------  --------  --------  --------  --------  --------

  Revision history:
    TSearchAndReplace
     Rev 01 - Control created.
     Rev 02 - SearchMemo() rewritten using ideas from a Borland DemoApp.
     Rev 03 - Full national chars support (runtime detect) added.
     Rev 04 - Minor bug-fixes implemented.

  --------  --------  --------  --------  --------  --------  --------  --------

  Thanks to:
    * Warny B Saurbrey - Who patiently Alfa-tests a lot of my stuff

  --------  --------  --------  --------  --------  --------  --------  --------

  TSearchAndReplace Properties:
    * Memo      - The TMemo that should be searced.
    * NotFound  - The message shown when no more matches is found. Should
                  contain "%s" for inserting the find-text.

  TSearchAndReplace Methods:
    * FindDialog    - Start a search session.
    * FindNext      - Continues a previous search
    * ReplaceDialog - Start a search and replace session

  --------  --------  --------  --------  --------  --------  --------  --------

  Known bugs and other issues:

    TSearchAndReplace
    * AV occured when memo contained more than 8 Kb text - fixed.
    * Memo didn't scroll correctly when text was found - fixed.

  --------  --------  --------  --------  --------  --------  --------  --------

  How to use:
    Drop a SearchAndReplace component on a form or datamodule. Link
    it to a Memo-field. Set the Memo's Hideselection property to false.
    Now just call the methods "FindDialog", "FindNext" and/or "ReplaceDialog"
    from your application's menu or toolbar. Nothing else is required.

  --------  --------  --------  --------  --------  --------  --------  --------
*)

unit SAR;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  stdCtrls;

type
  TDelimiter = Set of char;

type
  TSearchAndReplace = class(TComponent)
  private
    fFindDialog    : TFindDialog;
    fReplaceDialog : TReplaceDialog;
    fMemo          : TCustomEdit;
    fNotFound      : string;
  protected
    function  SearchMemo(Memo: TCustomEdit;Find:TFindDialog):Boolean;
    procedure OnFindEvent(sender:TObject);
    procedure OnReplaceEvent(sender:TObject);

  public
    WordDelimiters: TDelimiter;

    constructor Create(AnOwner:TComponent); override;
    destructor  Destroy; override;
    procedure   Free;

    procedure   FindDialog;
    procedure   ReplaceDialog;
    procedure   FindNext;
  published
    property  Memo    : TCustomEdit read fMemo write fMemo;
    property  NotFound: string read fNotFound write fNotFound;
  end;

procedure Register;

implementation

constructor TSearchAndReplace.Create(AnOwner:TComponent);
var
  c : char;
begin
  inherited ;
  Memo   :=NIL;
  fFindDialog   :=TFindDialog.Create(Application.MainForm);
  fReplaceDialog:=TReplaceDialog.Create(Application.MainForm);

  fFindDialog   .OnFind:=OnFindEvent;
  fReplaceDialog.OnFind:=OnFindEvent;
  fReplaceDialog.OnReplace:=OnReplaceEvent;
  fNotFound:='"%s" was not found in text';
  WordDelimiters:=[#0..#255];
  For c:=#0 to #255 do if isCharAlphaNumeric(c) then
    WordDelimiters:=WordDelimiters-[c];
end;

destructor TSearchAndReplace.Destroy;
begin
  fFindDialog.Destroy;
  fReplaceDialog.Destroy;
  inherited ;
end;

procedure TSearchAndReplace.free;
begin
  fFindDialog.free;
  fReplaceDialog.free;
  inherited ;
end;

function TSearchAndReplace.SearchMemo(Memo: TCustomEdit;Find:TFindDialog):boolean;

  function SearchBuf(const Buf: PChar; BufLen: Integer;
                     SelStart, SelLength: Integer;
                     const SearchString: String;
                     Options: TFindOptions): PChar;
  { This sub-function is a minor rewrite of a function in a Borland
    Demo-application.
  }
  var
    SearchCount, I: Integer;
    C: Char;
    Direction: integer;

    function FindNextWordStart(var BufPtr: PChar): Boolean;
    begin
      Result := False;
       { When Direction is forward (1), skip non delimiters, then skip delimiters. }
       { When Direction is backward (-1), skip delims, then skip non delims }
      while (SearchCount > 0) and
            ((Direction = 1) xor (BufPtr^ in WordDelimiters)) do begin
        Inc(BufPtr, Direction);
        Dec(SearchCount);
      end;
      while (SearchCount > 0) and
            ((Direction = -1) xor (BufPtr^ in WordDelimiters)) do begin
        Inc(BufPtr, Direction);
        Dec(SearchCount);
      end;
      Result := SearchCount >= 0;
      if (Direction = -1) and (BufPtr^ in WordDelimiters) then begin
        { back up one char, to leave ptr on first non delim }
        Dec(BufPtr, Direction);
        Inc(SearchCount);
      end;
    end;

  begin
    Result := nil;
    if BufLen <= 0 then Exit;
    if frDown in Options then begin
      Direction := 1;
      Inc(SelStart, SelLength);
      SearchCount := BufLen - SelStart - Length(SearchString);
      if SearchCount < 0 then Exit;
      if Longint(SelStart) + SearchCount > BufLen then Exit;
    end else begin
      Direction := -1;
      Dec(SelStart, Length(SearchString));
      SearchCount := SelStart;
    end;
    if (SelStart < 0) or (SelStart > BufLen) then Exit;
    Result := @Buf[SelStart];
    while SearchCount >= 0 do begin
      if frWholeWord in Options then
        if not FindNextWordStart(Result) then Break;
      I := 0;
      while (Result[I] = SearchString[I+1]) do begin
        Inc(I);
        if I >= Length(SearchString) then begin
          if (not (frWholeWord in Options)) or
             (SearchCount = 0) or
             (Result[I] in WordDelimiters) then Exit;
          Break;
        end;
      end;
      Inc(Result, Direction);
      Dec(SearchCount);
    end;
    Result := nil;
  end;
var
  size   : integer;
  pM,pO  : PChar;
  Buffer : pChar;
begin
  Result := False;
  if (Length(Find.FindText) = 0) or
     (Memo.GetTextLen = 0) then Exit;
  If frMatchCase in Find.Options then begin
    pO:=pChar(Memo.Text);
    pM:=SearchBuf(pO, Memo.GetTextLen, Memo.SelStart, Memo.SelLength,
                  Find.FindText, Find.Options);
  end else try
    // Fejl ved Search and Replace i filer over 8 Kb opstr muligvis her !
    // Kan det tnkes, at der kun er "plads" til 8K p heap/stack ?
    size:=Memo.GetTextLen;
    GetMem(Buffer,size+4);
    StrPLCopy(Buffer,AnsiLowercase(Memo.Text)+#0,Size+1);
    pO:=Buffer;
    pM:=SearchBuf(pO, size, Memo.SelStart, Memo.SelLength,
                  AnsiLowercase(Find.FindText), Find.Options);
  finally
    FreeMem(Buffer,size+4);
  end;
  if pM<>nil then begin
    Memo.SelStart:=pM-pO;
    Memo.SelLength:=Length(Find.FindText);
    Result := True;
  end;
end;

procedure TSearchAndReplace.OnFindEvent(sender:TObject);
begin
  If not SearchMemo(fMemo,TFindDialog(Sender)) then
   application.messagebox(
    pChar(Format(NotFound,[TFindDialog(Sender).FindText])),
    pChar(Application.Title), mb_ok or Mb_IconInformation);
  Memo.perform(EM_SCROLLCARET,0,0)
end;

procedure TSearchAndReplace.OnReplaceEvent(sender:TObject);
begin
  With TReplaceDialog(sender) do if not (frReplaceAll in options) then begin
    if fMemo.SelLength>0 then
      fMemo.SelText:=fReplaceDialog.ReplaceText;
    if not SearchMemo(fMemo,fReplaceDialog) then
      application.messagebox(
       pChar(Format(NotFound,[fReplaceDialog.FindText])),
       pChar(Application.Title),mb_ok or Mb_IconInformation);
  end else begin
    fMemo.SelStart:=0;
    fMemo.Sellength:=0;
    while SearchMemo(fMemo,fReplaceDialog) do
      fMemo.SelText:=fReplaceDialog.ReplaceText;
    fMemo.SelStart:=0;
  end;
end;

procedure TSearchAndReplace.FindDialog;
{ Shows the FindDialog and thus starts the find-session.
}
begin
  if fMemo.SelLength>0 then
    fFindDialog.FindText:=fMemo.Seltext;
  fFindDialog.Options:=[frDown];
  fMemo.SelLength:=0;
  fFindDialog.Execute;
end;

procedure TSearchAndReplace.FindNext;
{ Allows the search to be continued after the FindDialog is closed - in
  Windows Notepad this would be the user pressing "F3"
}
begin
  if fFindDialog.FindText>'' then
    OnFindEvent(fFindDialog);
  Memo.perform(EM_SCROLLCARET,0,0)
end;

procedure TSearchAndReplace.ReplaceDialog;
{ Shows the ReplaceDialog and starts a "Search & Replace" session
}
begin
  if fMemo.SelLength>0 then
    fReplaceDialog.FindText:=fMemo.Seltext;
  fReplaceDialog.Options:=[frDown];
  fMemo.SelLength:=0;
  fReplaceDialog.Execute;
end;

{ -- Register ---------------------------------------------------------------- }

procedure Register;
begin
  RegisterComponents('Samples', [TSearchAndReplace]);
end;

end.

(*
{ Search for textstring in a memo - may be faster than
  SearchMemo, but searches only forwards and cannot search
  for "whole word only"...
}
function SearchForward(Memo:TMemo;Find:TFindDialog):Boolean;
var
  pc   : pChar;
  fc   : pChar;
begin
  // Match Case
  With Find do if (frMatchcase in options) then begin
    pC :=pChar(memo.text);
    fc:=StrPos(@pC[memo.selstart+memo.sellength],pChar(findtext));
  end else begin
    pC :=pChar(AnsiLowercase(memo.text));
    fc:=StrPos(@pC[memo.selstart+memo.sellength],
               pChar(AnsiLowercase(findtext)));
  end;
  // Search
  if fc<>nil then with Find do begin
    memo.selstart:=fc-pc;
    memo.sellength:=length(findtext);
    memo.perform(EM_SCROLLCARET,0,0);
    result:=TRUE;
  end else
    result:=FALSE;
end;

*)
